import sys

from django.core.management import CommandError
from django.db import transaction

from ._private import BaseLiquidbCommand
from ...models import Snapshot, MigrationState


class Command(BaseLiquidbCommand):
    help = 'Create snapshot of migration state.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--name',
            type=str,
            help='Verbose name of snapshot',
        )
        parser.add_argument(
            '--overwrite',
            type=int,
            default=0,
            choices=[0, 1],
            help='If snapshot with given name exist it will overwrite it',
        )

    def _handle(self, *args, **options):
        snapshot_name = options['name']
        overwrite = bool(options['overwrite'])
        exists = Snapshot.objects.filter(name=snapshot_name).exists()
        if exists and not overwrite:
            raise CommandError(
                'Snapshot with given name already exists.\n'
                'Please use overwrite to execute it anyway',
            )

        if exists:
            snapshot = Snapshot.objects.get(name=snapshot_name)
        else:
            snapshot = Snapshot(name=snapshot_name)
        latest = Snapshot.objects.filter(applied=True).last()
        if latest is not None and latest.consistent_state:
            self.stdout.write('All migrations saved in currently applied snapshot. Nothing to create')
            sys.exit(0)
        with transaction.atomic():
            if latest is not None:
                latest.applied = False
                latest.save(update_fields=['applied'])
            migrations = self._latest_migrations().only('id', 'app', 'name')
            snapshot.applied = True

            if snapshot.id is not None:
                # in overwrite mode
                # so we need to delete all previous migrations
                # from this snapshot
                snapshot.migrations.all().delete()
            snapshot.save()
            to_create = [
                MigrationState(
                    snapshot=snapshot,
                    migration_id=migration.id,
                    app=migration.app,
                    name=migration.name,
                )
                for migration in migrations
            ]
            MigrationState.objects.bulk_create(to_create, batch_size=1000, ignore_conflicts=True)
            self.stdout.write(f'Snapshot "{snapshot_name}" successfully save')



import pytest
from liquidb.models import get_latest_applied_migrations_qs

_EXPECTED_APPS = [
    ('first_app', '0001'),
    ('second_app', '0003'),
    ('third_app', '0005'),
]


@pytest.fixture(scope='function')
def _create_apps_fixture(create_migration_state_fixture):
    create_migration_state_fixture(_EXPECTED_APPS)


@pytest.mark.django_db
@pytest.mark.usefixtures('_create_apps_fixture')
def test_latest_applied_migrations_qs():
    ids = get_latest_applied_migrations_qs()
    assert ids.count() == 3
    assert set(ids.values_list('app', flat=True)) == set([app_name for app_name, _ in _EXPECTED_APPS])
    assert set(ids.values_list('name', flat=True)) == set([migration_name for _, migration_name in _EXPECTED_APPS])



import pytest

from liquidb.models import _create_migration_hash


@pytest.mark.parametrize(
    'state_1, state_2,expected',
    [
        (
                [('first_app', '0001'), ('second_app', '0002')],
                [('second_app', '0002'), ('first_app', '0001')],
                True,
        ),
        (
                [],
                [],
                True,
        ),
        (
                [('first_app', '0001'), ('second_app', '0002')],
                [('first_app', '0001'), ('second_app', '0002')],
                True,
        ),
        (
                [('first_app', '0001'), ('second_app', '0002'), ('third_app', '0003')],
                [('first_app', '0001'), ('second_app', '0002')],
                False,
        ),
        (
                [('first_app', '0001'), ('second_app', '0002')],
                [('second_app', '0002')],
                False,
        ),
        (
                [('first_app', '0001'), ('second_app', '0002')],
                [],
                False,
        ),
        (
                [('second_app', '0002')],
                [('first_app', '0001')],
                False,
        ),
    ]
)
def test_hash_function(state_1, state_2, expected):
    s1 = _create_migration_hash(state_1)
    s2 = _create_migration_hash(state_2)
    result = s1 == s2
    assert result is expected





# Generated by Django 3.1.6 on 2021-03-04 08:49

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import liquidb.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='MigrationState',
            fields=[
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False, unique=True)),
                ('migration_id', models.IntegerField(db_index=True)),
                ('app', models.CharField(max_length=255)),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Snapshot',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField(db_index=True, default=liquidb.models._generate_commit_name, unique=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('applied', models.BooleanField(default=False)),
            ],
        ),
        migrations.AddIndex(
            model_name='snapshot',
            index=models.Index(condition=models.Q(applied=True), fields=['applied'], name='unique_applied'),
        ),
        migrations.AddField(
            model_name='migrationstate',
            name='snapshot',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='migrations', to='liquidb.snapshot'),
        ),
    ]



import pytest
from django.core.management import call_command

from liquidb.models import Snapshot


@pytest.mark.django_db
@pytest.mark.parametrize(
    'apps', [
        pytest.param(
            [('first_app', '0001'), ('second_app', '0002'), ('third_app', '0003')],
            id='Few apps '
        ),
        pytest.param(
            [],
            id='Empty migration state',
        )
    ]
)
def test_delete_snapshot_history(apps, create_migration_state_fixture):
    create_migration_state_fixture(apps)
    call_command('create_migration_snapshot', name='init')
    call_command('delete_snapshot_history', interactive=False)
    assert Snapshot.objects.count() == 0



from django.db import connection, transaction
from django.db.migrations.recorder import MigrationRecorder


def change_state_mock(self, snapshot):  # pylint: disable=unused-argument
    # unfortunately no other way that we can chane state in tests
    # without all migrations files and whole django machinery
    # so for this test we will mock this state
    recorder = MigrationRecorder(connection=connection)
    Migration = recorder.Migration
    migration_state = snapshot.migrations.values_list('app', 'name')
    with transaction.atomic():
        Migration.objects.all().delete()
        for app, name in migration_state:
            m = Migration(
                app=app,
                name=name,
            )
            m.save()


import sys

from django.core.management import CommandError
from django.db import transaction

from ._private import BaseLiquidbRevertCommand


class Command(BaseLiquidbRevertCommand):
    help = 'Delete snapshot by name'

    def add_arguments(self, parser):
        parser.add_argument(
            '--name',
            type=str,
            help='Name of snapshot to delete',
        )
        parser.add_argument(
            '--noinput',
            action='store_false',
            dest='interactive',
            default=True,
            help='Do not ask for confirmation '
        )

    def _handle(self, *args, **options):
        name = options['name']
        interactive = options['interactive']
        snapshot = self.get_snapshot(name)
        if interactive:
            confirm = input(f"""
                You have requested to delete snapshot "{name}" with all migrations connected to it.
                This action is IRREVERSIBLE. 
                Are you sure you want to do this?
    
                Type 'yes' to continue, or 'no' to cancel: """
            )
            if confirm != 'yes':
                self.stdout.write('Snapshot deletion is canceled')
                sys.exit(0)

        if snapshot.applied:
            raise CommandError(f'Snapshot with name {name} is applied and couldn\'t be deleted')
        with transaction.atomic():
            _total, models = snapshot.delete()
            migrations_states = models.get('liquidb.MigrationState', 0)

        self.stdout.write(f'Successfully deleted snapshot "{name}" and {migrations_states} migrations')



__all__ = [
    'CheckoutCommand',
    'CheckoutLatestCommand',
    'SnapshotCommand',
]
from ..management.commands.checkout_snapshot import Command as CheckoutCommand
from ..management.commands.checkout_latest_snapshot import Command as CheckoutLatestCommand
from ..management.commands.create_migration_snapshot import Command as SnapshotCommand



SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'tests',
    'liquidb',
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    }
}
USE_TZ = True



from ._private import BaseLiquidbRevertCommand


class Command(BaseLiquidbRevertCommand):
    help = 'Revert migration state to given commit (snapshot)'

    def add_arguments(self, parser):
        parser.add_argument(
            '--name',
            type=str,
            help='Name of snapshot to apply',
        )
        parser.add_argument(
            '--force',
            type=int,
            default=0,
            choices=[0, 1],
            help='Drop unsaved changes',
        )

    def _handle(self, *args, **options):
        snapshot = self.get_snapshot(options['name'])
        self._checkout_snapshot(snapshot, force=bool(options['force']))



import sys

from django.db import transaction

from ._private import BaseLiquidbRevertCommand
from ...models import Snapshot


class Command(BaseLiquidbRevertCommand):
    help = 'Delete all history of snapshots'

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput',
            action='store_false',
            dest='interactive',
            default=True,
            help='Do not ask for confirmation '
        )

    def _handle(self, *args, **options):
        interactive = options['interactive']
        if interactive:
            confirm = input(f"""
                You have requested to delete snapshot history with all migrations.
                This action is IRREVERSIBLE. 
                Are you sure you want to do this?

                Type 'yes' to continue, or 'no' to cancel: """
                            )
            if confirm != 'yes':
                self.stdout.write('Snapshot history deletion is canceled')
                sys.exit(0)
        with transaction.atomic():
            _total, models = Snapshot.objects.all().delete()
            snapshots = models.get('liquidb.Snapshot', 0)
            migrations_states = models.get('liquidb.MigrationState', 0)
        self.stdout.write(f'Successfully deleted history of {snapshots} snapshots and {migrations_states} migrations')



from django.core.exceptions import ObjectDoesNotExist
from django.core.management import CommandError

from ._private import BaseLiquidbRevertCommand
from ...models import Snapshot


class Command(BaseLiquidbRevertCommand):
    help = 'Revert migration to latest commit (snapshot)'

    def add_arguments(self, parser):
        parser.add_argument(
            '--force',
            type=int,
            default=0,
            choices=[0, 1],
            help='Drop unsaved changes',
        )

    def _handle(self, *args, **options):
        try:
            snapshot = Snapshot.objects.latest('id')
        except ObjectDoesNotExist as error:
            raise CommandError('No latest snapshot present') from error
        self._checkout_snapshot(snapshot, force=bool(options['force']))



from unittest.mock import patch

import pytest
from django.core.management import call_command
from liquidb.management.commands._private import BaseLiquidbRevertCommand  # NOQA # pylint: disable=protected-access
from liquidb.models import Snapshot
from tests.tools_tests import change_state_mock

_STATE_TEMPLATE = 'state_{}'


@pytest.fixture(scope='function')
def _create_three_state_fixture(create_migration_state_fixture):
    state_template = [('first_app', '000{}'), ('second_app', '000{}'), ('third_app', '000{}')]
    for counter in range(1, 4):
        state = [
            (app_name, migration.format(counter))
            for app_name, migration in state_template
        ]
        create_migration_state_fixture(state)
        call_command('create_migration_snapshot', name=_STATE_TEMPLATE.format(counter))


@pytest.mark.django_db
@patch.object(
    BaseLiquidbRevertCommand,
    '_revert_to_snapshot',
    new=change_state_mock,
)
def test_checkout_between_snapshots(_create_three_state_fixture):
    state_1 = _STATE_TEMPLATE.format(1)
    state_2 = _STATE_TEMPLATE.format(2)
    state_3 = _STATE_TEMPLATE.format(3)
    call_command('checkout_snapshot', name=state_1)
    snapshot = Snapshot.objects.get(name=state_1)
    assert snapshot.consistent_state is True
    # state_3
    call_command('checkout_snapshot', name=state_3)
    snapshot = Snapshot.objects.get(name=state_3)
    assert snapshot.consistent_state is True
    # state_2
    call_command('checkout_snapshot', name=state_2)
    snapshot = Snapshot.objects.get(name=state_2)
    assert snapshot.consistent_state is True



from io import StringIO

import pytest
from django.core.management import call_command, CommandError
from mock import patch

from liquidb.models import Snapshot


@pytest.mark.django_db
@pytest.mark.parametrize(
    'apps', [
        pytest.param(
            [('first_app', '0001'), ('second_app', '0003'), ('third_app', '0005')],
            id='Three apps'
        ),
        pytest.param(
            [('first_app', '0001'), ('second_app', '0005')],
            id='Two apps'
        ),
        pytest.param(
            [('first_app', '0001')],
            id='Single app'
        ),
        pytest.param(
            [],
            id='Empty',
        )
    ]
)
def test_create_snapshot(apps, create_migration_state_fixture):
    expected = len(apps)
    create_migration_state_fixture(apps)
    s_name = 'init'
    call_command('create_migration_snapshot', name=s_name)
    assert Snapshot.objects.count() == 1
    snapshot = Snapshot.objects.first()
    assert snapshot.name == s_name
    assert snapshot.migrations.count() == expected


@pytest.mark.django_db
@pytest.mark.parametrize(
    'overwrite', [0, 1]
)
def test_snapshot_with_same_name_state_is_not_changed(create_migration_state_fixture, overwrite):
    apps = [('first_app', '0001'), ('second_app', '0005')]
    snapshot_name = 'first'
    create_migration_state_fixture(apps)
    call_command('create_migration_snapshot', name=snapshot_name)
    if not overwrite:
        with pytest.raises(CommandError):
            call_command('create_migration_snapshot', name=snapshot_name, overwrite=overwrite)
    else:
        with pytest.raises(SystemExit):
            call_command('create_migration_snapshot', name=snapshot_name, overwrite=overwrite)


@pytest.mark.django_db
@pytest.mark.parametrize(
    'overwrite', [0, 1]
)
def test_snapshot_with_same_name_state_is_changed(create_migration_state_fixture, overwrite):
    snapshot_name = 'stable'
    state_1 = [('first_app', '0001'), ('second_app', '0005')]
    create_migration_state_fixture(state_1)
    call_command('create_migration_snapshot', name=snapshot_name)
    state_2 = [('first_app', '0002'), ('second_app', '0003'), ('third_app', '0006')]
    create_migration_state_fixture(state_2)
    if not overwrite:
        with pytest.raises(CommandError):
            call_command('create_migration_snapshot', name=snapshot_name, overwrite=overwrite)
    else:
        call_command('create_migration_snapshot', name=snapshot_name, overwrite=overwrite)
        snapshot = Snapshot.objects.get(name=snapshot_name)
        # check that snapshot overwritten
        assert snapshot.consistent_state is True
        expected_state = state_2
        assert snapshot.migrations.count() == len(expected_state)
        migrations = snapshot.migrations.all()
        assert set(migrations.values_list('app', flat=True)) == set([app_name for app_name, _ in expected_state])
        assert set(migrations.values_list('name', flat=True)) == set(
            [migration_name for _, migration_name in expected_state]
        )


@pytest.mark.django_db
@patch('sys.stdout', new_callable=StringIO)
def test_snapshot_with_same_hash(patched_stdout, create_migration_state_fixture):
    apps = [('first_app', '0001'), ('second_app', '0005')]
    create_migration_state_fixture(apps)
    call_command('create_migration_snapshot', name='first')
    with pytest.raises(SystemExit):
        call_command('create_migration_snapshot', name='second')
    patched_stdout.assert_called_once_with('All migrations saved in currently applied snapshot. Nothing to create')


@pytest.mark.django_db
@patch('sys.stderr')
def test_snapshot_with_same_hash(patched_stderr, create_migration_state_fixture):
    first_state = [('first_app', '0001'), ('second_app', '0005')]
    create_migration_state_fixture(first_state)
    call_command('create_migration_snapshot', name='first')
    second_state = [('first_app', '0003'), ('third_app', '0001'), ('second_app', '0006')]
    create_migration_state_fixture(second_state)
    patched_stderr.assert_not_called()



import pytest
from django.core.management import call_command, CommandError

from liquidb.models import Snapshot


@pytest.mark.django_db
def test_delete_snapshot_by_name_applied(create_migration_state_fixture):
    apps = [('first_app', '0001'), ('second_app', '0002'), ('third_app', '0003')]
    create_migration_state_fixture(apps)
    snapshot_name = 'init'
    call_command('create_migration_snapshot', name=snapshot_name)
    with pytest.raises(CommandError):
        call_command('delete_snapshot_by_name', name=snapshot_name, interactive=False)
    assert Snapshot.objects.count() == 1


@pytest.mark.django_db
def test_delete_snapshot_by_name_not_applied(create_migration_state_fixture):
    apps = [('first_app', '0001'), ('second_app', '0002'), ('third_app', '0003')]
    create_migration_state_fixture(apps)
    snapshot_name = 'first'
    call_command('create_migration_snapshot', name=snapshot_name)
    apps = [('first_app', '0002'), ('second_app', '0003'), ('third_app', '0004')]
    create_migration_state_fixture(apps)
    call_command('create_migration_snapshot', name='second')
    call_command('delete_snapshot_by_name', name=snapshot_name, interactive=False)
    assert Snapshot.objects.count() == 1
    snapshot = Snapshot.objects.first()
    assert snapshot.applied is True
    assert snapshot.consistent_state is True



import sys
from abc import ABCMeta, abstractmethod

from django.core.exceptions import ObjectDoesNotExist
from django.core.management import BaseCommand, CommandError
from django.db import connection, transaction
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.recorder import MigrationRecorder
from django.db.migrations.state import ProjectState

from ...models import Snapshot, MigrationState, get_latest_applied_migrations_qs


class BaseLiquidbCommand(BaseCommand, metaclass=ABCMeta):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connection = connection

    def _check_migration_db_exists(self):
        recorder = MigrationRecorder(self.connection)
        if not recorder.has_table():
            raise CommandError('No migration table present')

    def _check_liquidb_tables_exists(self):
        with self.connection.cursor() as cursor:
            tables = self.connection.introspection.table_names(cursor)
        tables_to_check = [Snapshot._meta.db_table, MigrationState._meta.db_table]  # NOQA
        if tables and tables_to_check:
            return all(map(lambda x: x in tables, tables_to_check))
        return False

    def _init(self):
        self._check_migration_db_exists()
        self._check_liquidb_tables_exists()

    def handle(self, *args, **options):
        self._init()
        self._handle(*args, **options)

    def _latest_migrations(self):
        return get_latest_applied_migrations_qs(self.connection)

    @staticmethod
    def get_snapshot(name: str) -> Snapshot:
        """Retrun snapshot by name"""
        try:
            snapshot = Snapshot.objects.get(name=name)
        except ObjectDoesNotExist as error:
            raise CommandError(f'Snapshot with name: "{name}" doesn\'t exists') from error
        return snapshot

    @abstractmethod
    def _handle(self, *args, **options):
        raise NotImplementedError


class BaseLiquidbRevertCommand(BaseLiquidbCommand, metaclass=ABCMeta):

    @staticmethod
    def _get_applied_snapshot():
        """Return latest applied snapshot by id"""
        try:
            latest = Snapshot.objects.filter(applied=True).latest('id')
        except ObjectDoesNotExist as error:
            raise CommandError('No latest snapshot present') from error
        return latest

    def _revert_to_snapshot(self, snapshot: Snapshot):
        """Revert db state to given snapshot"""
        targets = snapshot.migrations.values_list('app', 'name')
        if not targets:
            # snapshot that do not have any migrations
            # TODO get all apps from django apps and migrate everything to zero ?
            raise CommandError(
                f'No connected migrations found for snapshot {snapshot.name}'
            )
        executor = MigrationExecutor(self.connection)
        try:
            _: ProjectState = executor.migrate(targets)
        except KeyError:
            message = (
                '\n'.join(
                    [
                        f'* App Name: {app} ; Migration: {migration}'
                        for app, migration in targets
                    ]
                )
            )
            raise CommandError(
                f'\nSome migrations are missing.\n'
                f'Please make sure all migrations in snapshot: "{snapshot.name}";\n'
                f'Are present in corresponding apps: \n{message}'
            )

    def _checkout_snapshot(self, snapshot: Snapshot, force=False):
        """Run all checks before reverting to desired state"""
        latest = self._get_applied_snapshot()
        consistent_with_migration_table = latest.consistent_state
        if not consistent_with_migration_table and not force:
            raise CommandError(
                'Migration state is inconsistent latest applied '
                'snapshot do not has all currently applied migrations. '
                'Please use --force flag to checkout anyway.'
            )
        if not consistent_with_migration_table:
            self.stdout.write('Force to apply migration. Unsaved changes will be dropped')

        # eq is not same as equal snapshot.id == latest.id
        # see liquidb.models.Snapshot.__eq__
        if snapshot == latest:
            # migration hashes are the same
            # nothing to do
            self.stdout.write(f'Snapshot "{latest.name}" already applied ')
            sys.exit(0)
        self._revert_to_snapshot(snapshot)
        with transaction.atomic():
            latest.applied = False
            latest.save(update_fields=['applied'])

            snapshot.applied = True
            snapshot.save(update_fields=['applied'])

        self.stdout.write(f'Checkout from snapshot "{latest.name}" to "{snapshot.name}"')

    @abstractmethod
    def _handle(self, *args, **options):
        raise NotImplementedError
